     /**
      * \brief Flat File Parser : Grouping Line Processor
      *
      * This line processor is grouping the data passed from the parser to the
      * line processor. The start of a block/group of data is determined by the
      * start line format. The end of a block/group of data is determined by the
      * end line format. Those line formats can be a fixed or variable line 
      * format. They may be set on or after instance creation.
      * <p>
      * After the end of a group of data has been reached the registered 
      * program will be called with the grouped data. This program will 
      * make the custom processing of the data.
      * <p>
      * If the processing program (executor) is throwing an escape message
      * the error will be caught, logged and the data serialized.
      * <p>
      * If one or more lines are invalid the group of data won't be passed to
      * the executor but serialized with the registered serializer. 
      * <p>
      * The executor program must implement the following parameter list:
      * <ul>
      *   <li>logger - pointer (const)</li>
      *   <li>data - pointer (const)</li>
      * </ul>
      * The data is a linked list of data where each entry is a data structure
      * (see ffp_processor_grouping_line). 
      * 
      * \info For this to work the parser must be configured to also pass invalid
      *       lines to the line processor. Those lines are flagged as invalid.
      * 
      * \author Mihael Schmidt
      * \date   31.12.2010
      *
      * \implements FFPPRC - Flat File Parser Line Processor
      *
      */

      *-------------------------------------------------------------------------
      *
      * (C) Copyleft 2011 Mihael Schmidt
      *
      * This file is part of Flat File Parser (FFP).
      *
      * FFP is free software: you can redistribute it and/or modify
      * it under the terms of the GNU General Public License as published
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * FFP is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU General Public License for more details.
      *
      * You should have received a copy of the GNU General Public License
      * along with FFP.  If not, see <http://www.gnu.org/licenses/>.
      *
      *-------------------------------------------------------------------------

     H nomain


      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D processCollectedData...
     D                 PR
     D   processor                     *   const
      *
      /copy 'parser_h.rpgle'
      /copy 'line_processor_h.rpgle'
      /copy 'grouping_line_processor_h.rpgle'
      /copy 'logger_h.rpgle'
      /copy 'serializer_h.rpgle'
      /copy 'utilities_h.rpgle'
      /copy 'message_h.rpgle'
      /copy 'llist_h.rpgle'
      /copy 'libc_h.rpgle'


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
     D LINE_PROCESSOR_ID...
     D                 C                   'ffp_processor_grouping'


      *---------------------------------------------------------------
      * Templates
      *---------------------------------------------------------------
     D tmpl_userdata...
     D                 DS                  template qualified
     D   lineFormats                       likeds(tmpl_lineFormats)
     D   lines                         *
     D   errorLines                    *
     D   executor                    10A
      *
     D tmpl_lineFormats...
     D                 DS                  template qualified
     D   start                         *
     D   end                           *
      *
      /include 'line_processor_t.rpgle'
      /include 'program_data_structure_h.rpgle'


      *---------------------------------------------------------------
      * Procedures
      *---------------------------------------------------------------

     /**
      * \brief Create Grouping Line Processor
      *
      * Creates a grouping line processor instance. The starting and ending 
      * line formats and the executing program can either be passed as parameters
      * or  
      *
      * \param Benutzerdaten
      */
     P ffp_processor_grouping_create...
     P                 B                   export
     D                 PI              *
     D   parser                        *   const 
     D   pUserdata                     *   const options(*nopass)
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(lineProcessor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
     D parameter       S            100A
     D ptr             S               *
      /free
       lineProcessor = %alloc(%size(ffp_processor));

       clear header;

       header.id = LINE_PROCESSOR_ID;
       header.logger = *null;
       header.proc_process = %paddr('ffp_processor_grouping_process');
       header.proc_finalize = %paddr('ffp_processor_grouping_finalize');
       header.proc_setLogger = %paddr('ffp_processor_grouping_setLogger');
       header.proc_setSerializer=%paddr('ffp_processor_grouping_setSerializer');

       header.userdata = %alloc(%size(tmpl_userdata));
       userdata.lines = list_create();
       userdata.lineformats.start = *null;
       userdata.lineformats.end = *null;
       
       // errorLines is a linked list. 
       // Each entry is an integer which is the index of the invalid line in the
       // data list. The index is zero-based.
       userdata.errorLines = list_create();
       
       userdata.executor = *blank;

       if (%parms() = 2 and pUserdata <> *null);
         ptr = strtok(pUserdata : ' ');
         if (ptr <> *null);
           parameter = %str(ptr);
           ffp_processor_grouping_setStartLineFormat(
               lineProcessor : 
               ffp_getLineFormat(parser : %trimr(parameter)));
         endif;
       
         ptr = strtok(*null : ' ');
         if (ptr <> *null);
           parameter = %str(ptr);
           ffp_processor_grouping_setEndLineFormat(
               lineProcessor : 
               ffp_getLineFormat(parser : %trimr(parameter)));
         endif;
       
         ptr = strtok(*null : ' ');
         if (ptr <> *null);
           parameter = %str(ptr);
           ffp_processor_grouping_setExecutor(lineProcessor : parameter);
         endif;
       endif;

       return lineProcessor;
      /end-free
     P                 E


     /**
      * \brief Dispose line processor
      *
      * Frees all allocated memory.
      *
      * \param Line processor
      */
     P ffp_processor_grouping_finalize...
     P                 B                   export
     D                 PI
     D   lineProcessor...
     D                                 *
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(lineProcessor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
     D command         S            100A
      /free
       list_dispose(userdata.lines);
       list_dispose(userdata.errorLines);

       if (header.userdata <> *null);
         dealloc header.userdata;
       endif;

       dealloc(n) lineProcessor;
      /end-free
     P                 E


     /**
      * \brief Process line data
      *
      * Group data here and process it with the given program.
      * Invalid data will be passed to the executing program depending on the
      * parser configuration.
      *
      * \param Line processor
      * \parma Parser
      * \param Event
      * \param Line format
      * \param Data
      */
     P ffp_processor_grouping_process...
     P                 B                   export
     D                 PI
     D  lineProcessor                  *   const
     D  parser                         *   const
     D  event                        10I 0 const
     D  lineFormatPtr                  *   const
     D  data                      65535A   const varying
     D  valid                          N   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(lineProcessor)
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
     D errorLineNumber...
     D                 S             10I 0
     D lineEntry       DS                  likeds(ffp_processor_grouping_line)
     D serTransId      S           1000A   varying
      /free
       if (event = LINE);

         lineEntry.lineFormat = lineFormatPtr;
         lineEntry.data = data;
         lineEntry.lineNumber = ffp_getPhysicalLineNumber(parser);
         list_add(userdata.lines : %addr(lineEntry) : %size(lineEntry));

         if (not valid);
           errorLineNumber = list_size(userdata.lines) -1;
           list_addInteger(userdata.errorLines : errorLineNumber);
         endif;

         if (lineFormatPtr = userdata.lineFormats.end);
           
           processCollectedData(lineProcessor);

           // reset line buffer and error buffer/state
           list_clear(userdata.lines);
           list_clear(userdata.errorLines);
         endif;

       elseif (event = DOCUMENT_START);
       
         // check if a starting and ending line format is set
         if (userdata.lineformats.start = *null);
           ffp_logger_log(header.logger : '[' + %trimr(header.id) + '] ' +
                          'No starting line format is registered on the ' + 
                          'grouping line processor.' : *omit : 
                          FFP_LOG_LEVEL_FATAL); 
         elseif (userdata.lineformats.end = *null);
           ffp_logger_log(header.logger : '[' + %trimr(header.id) + '] ' +
                          'No ending line format is registered on the ' + 
                          'grouping line processor.' : *omit : 
                          FFP_LOG_LEVEL_FATAL); 
         endif;

       endif;
      /end-free
     P                 E


     /**
      * \brief Set logger
      *
      * \param Processor
      * \param Logger
      */
     P ffp_processor_grouping_setLogger...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   logger                        *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
      /free
       header.logger = logger;
      /end-free
     P                 E


     /**
      * \brief Set serializer
      *
      * \param Processor
      * \param Serializer
      */
     P ffp_processor_grouping_setSerializer...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   serializer                    *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
      /free
       header.serializer = serializer;
      /end-free
     P                 E


     /**
      * \brief Set starting line format
      *
      * \param Processor
      * \param Line Format
      */
     P ffp_processor_grouping_setStartLineFormat...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   lineFormat                    *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
      /free
       userdata.lineFormats.start = lineFormat;
      /end-free
     P                 E


     /**
      * \brief Set ending line format
      *
      * \param Processor
      * \param Line Format
      */
     P ffp_processor_grouping_setEndLineFormat...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   lineFormat                    *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
      /free
       userdata.lineFormats.end = lineFormat;
      /end-free
     P                 E

 
     /**
      * \brief Set executor program
      * 
      * Sets the executor program. Any previous set programs will be overwritten.
      * 
      * \param Processor
      * \param Executor program name
      */ 
     P ffp_processor_grouping_setExecutor...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   executor                    10A   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
      /free
       userdata.executor = executor;
      /end-free
     P                 E  
       
     
     /**
      * \brief Get executor program
      * 
      * \param Processor
      * 
      * \return Executor program name or *blank if no program has been set
      */ 
     P ffp_processor_grouping_getExecutor...
     P                 B                   export
     D                 PI            10A 
     D   processor                     *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
      /free
       return userdata.executor;
      /end-free
     P                 E


     /**
      * \brief Process data group
      *
      * The grouped data will be passed to the configured program. That
      * program is the real processor of the data. 
      * <p>
      * If an error occurs the exception will be caught and the grouped data
      * serialized if a serializer has been registered at the parser.
      * <p>
      * If the grouped data has invalid lines the grouped data will be serialized
      * if a serializer has been registered at the parser.
      * <p>
      * No error/exception will be passed to the parser instance. So the parser
      * won't be able to determine the outcome of the processing by itself.
      *
      * \param Processor
      */
     P processCollectedData...
     P                 B
     D                 PI
     D   processor                     *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D userdata        DS                  likeds(tmpl_userdata)
     D                                     based(header.userdata)
      *
      *
     D processData     PR                  extpgm(userdata.executor)
     D   logger                        *   const
     D   data                          *   const
      *
      *
     D serTransId      S           1000A   varying
     D mde             S             15P 5
      /free
       monitor;
       
         if (userdata.executor = *blank);  // check if executor program has been registered
           msg_sendEscapeMessage('No executor program has been registered. ' +
                                 'The grouped data cannot be processed.');
                                 
         elseif (list_size(userdata.errorLines) = 0); // check for invalid lines
           processData(header.logger : userdata.lines);
         
           ffp_logger_log(header.logger :
                          '[' + %trimr(header.id) + '] Grouped data has been ' +
                          'processed by program ' + %trimr(userdata.executor) + 
                          '.' : *omit : FFP_LOG_LEVEL_ANNOYING);
                          
         else;
           msg_sendEscapeMessage('The grouped data cannot be processed due ' + 
                                 'to previous errors (see log).');
         endif;

         on-error *all;

           ffp_logger_log(header.logger : '[' + %trimr(header.id) + '] ' +
                          programStatusDs.messageId + ': ' +
                          ffp_util_getLastErrorMessage(
                              programStatusDs.messageId));
           if (header.serializer <> *null);
             serTransId = ffp_serializer_serialize(header.serializer :
                                                   userdata.lines);
             ffp_logger_log(header.logger :
                            '[' + %trimr(header.id) + '] ' +
                            'The data has been serialized.' : serTransId :
                            FFP_LOG_LEVEL_INFO);
           endif;
           
       endmon;
      /end-free
     P                 E
