     /**
      * \brief Flat File Parser : Line Processor Interface
      *
      *
      * \author Mihael Schmidt
      * \date   07.04.2010
      *
      * \interface
      *
      */

      *-------------------------------------------------------------------------
      *
      * (C) Copyleft 2011 Mihael Schmidt
      *
      * This file is part of Flat File Parser (FFP).
      *
      * FFP is free software: you can redistribute it and/or modify
      * it under the terms of the GNU General Public License as published
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * FFP is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU General Public License for more details.
      *
      * You should have received a copy of the GNU General Public License
      * along with FFP.  If not, see <http://www.gnu.org/licenses/>.
      *
      *-------------------------------------------------------------------------


     H nomain

      /define INTERFACE


      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D ffp_processor_check...
     D                 PR
     D   processor                     *   const
      *
      /copy 'line_processor_h.rpgle'
      /copy 'line_processor_t.rpgle'
      /copy 'message_h.rpgle'
      /copy 'reflection_h.rpgle'


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------


     /**
      * \brief Create line processor
      *
      * Creates the line processor.
      *
      * \param Service program (Line Processor)
      * \param User data (optional)
      * \param Create procedure (default: ffp_processor_create)
      *
      * \return Line Processor
      *
      * \throws CPF9898 Incomplete implementation
      */
     P ffp_processor_create...
     P                 B                   export
     D                 PI              *
     D   serviceProgram...
     D                               10A   const
     D   parser                        *   const 
     D   userdata                      *   const options(*nopass)
     D   createProcedure...
     D                              256A   const options(*nopass)
      *
     D procedurePointer...
     D                 S               *   procptr
     D create          PR              *   extproc(procedurePointer)
     D   parser                        *   const 
     D   userdata                      *   const options(*nopass)
      *
     D procedureName...
     D                 S            256A
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(lineProcessor)
      /free
       if (%parms() = 4);
         procedureName = createProcedure;
       else;
         procedureName = 'ffp_processor_create';
       endif;

       procedurePointer = reflection_getProcedurePointer(
                                  serviceProgram :
                                  %trimr(procedureName));

       if (procedurePointer <> *null);

           // line processor aufbauen/erstellen
           if (%parms() = 2);
             lineProcessor = create(parser);
           else;
             lineProcessor = create(parser : userData);
           endif;

       endif;

       ffp_processor_check(lineProcessor);

       return lineProcessor;
      /end-free
     P                 E


     /**
      * \brief Dispose line processor
      *
      * Calls clean up process and frees allocated memory.
      *
      * \param Line processor
      */
     P ffp_processor_finalize...
     P                 B                   export
     D                 PI
     D  processor                      *
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D procedurePointer...
     D                 S               *   procptr
     D finalize        PR                  extproc(procedurePointer)
     D   processor                     *
      /free
       procedurePointer = header.proc_finalize;

       ffp_processor_check(processor);

       finalize(processor);

       if (processor <> *null);
         dealloc(n) processor;
       endif;
      /end-free
     P                 E


     /**
      * \brief Process line
      *
      * Process the passed line of data.
      *
      * \param Line processor
      * \param Data
      */
     P ffp_processor_process...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   parser                        *   const
     D   event                       10I 0 const
     D   lineFormat                    *   const
     D   data                     65535A   const varying
     D   valid                         N   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
      *
     D process         PR                  extproc(header.proc_process)
     D   processor                     *   const
     D   parser                        *   const
     D   event                       10I 0 const
     D   lineFormat                    *   const
     D   data                     65535A   const varying
     D   valid                         N   const
      /free
       ffp_processor_check(processor);

       process(processor : parser : event : lineFormat : data : valid);
      /end-free
     P                 E


     /**
      * \brief Set logger
      *
      * \param Line processor
      * \param Logger
      */
     P ffp_processor_setLogger...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   logger                        *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
      *
     D setLogger...
     D                 PR                  extproc(header.proc_setLogger)
     D   processor                     *   const
     D   logger                        *   const
      /free
       ffp_processor_check(processor);

       setLogger(processor : logger);
      /end-free
     P                 E


     /**
      * \brief Set serializer
      *
      * \param Processor
      * \param Serializer
      */
     P ffp_processor_setSerializer...
     P                 B                   export
     D                 PI
     D   processor                     *   const
     D   serializer                    *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
      *
     D setSerializer...
     D                 PR                  extproc(header.proc_setSerializer)
     D   processor                     *   const
     D   serializer                    *   const
      /free
       ffp_processor_check(processor);

       setSerializer(processor : serializer);
      /end-free
     P                 E


     /**
      * \brief Check line processor
      *
      * Checks if the line processor provides procedure pointers for all
      * procedures it should implement.
      *
      * \param Line processor
      *
      * \throws CPF9898 Incomplete implementation
      */
     P ffp_processor_check...
     P                 B
     D                 PI
     D   processor                     *   const
      *
     D header          DS                  likeds(ffp_processor)
     D                                     based(processor)
     D cachedProviderId...
     D                 S                   like(ffp_processor.id) static
     D providerChecked...
     D                 S               N   inz(*off) static
      /free
       if (header.id = cachedProviderId and providerChecked);
         return;
       endif;

       cachedProviderId = header.id;

       if (header.proc_process = *null);
         msg_sendEscapeMessage('Flat File Parser Line Processor Interface: ' +
             'Incomplete Implementation ' + %trimr(header.id)+
             '.');
       elseif (header.proc_finalize = *null);
         msg_sendEscapeMessage('Flat File Parser Line Processor Interface: ' +
             'Incomplete Implementation ' + %trimr(header.id)+
             '.');
       elseif (header.proc_setLogger = *null);
         msg_sendEscapeMessage('Flat File Parser Line Processor Interface: ' +
             'Incomplete Implementation ' + %trimr(header.id)+
             '.');
       elseif (header.proc_setSerializer = *null);
         msg_sendEscapeMessage('Flat File Parser Line Processor Interface: ' +
             'Incomplete Implementation ' + %trimr(header.id)+
             '.');
       endif;

       providerChecked = *on;
      /end-free
     P                 E

