     /**
      * \brief Flat File Parser : Main Module
      *
      * <br><br>
      *
      * This software is inspired by the Java Flat File Parsing library from
      * dr. Cristiano Sadun (hosted on Sourceforge.net).
      *
      * \author Mihael Schmidt
      * \date   07.01.2008
      *
      * \link http://jffp.sourceforge.net Java Flat File Parsing library
      *
      */

      *-------------------------------------------------------------------------
      *
      * (C) Copyleft 2011 Mihael Schmidt
      *
      * This file is part of Flat File Parser (FFP).
      *
      * FFP is free software: you can redistribute it and/or modify
      * it under the terms of the GNU General Public License as published
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * FFP is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU General Public License for more details.
      *
      * You should have received a copy of the GNU General Public License
      * along with FFP.  If not, see <http://www.gnu.org/licenses/>.
      *
      *-------------------------------------------------------------------------


     H nomain


      *---------------------------------------------------------------
      * Constants
      *---------------------------------------------------------------
     D PROGRAM_ID      C                   'ffp_parser'
      *
     D NUMERIC         C                   '+-,.0123456789'


      *---------------------------------------------------------------
      * Prototypen
      *---------------------------------------------------------------
     D resolveMatchingLineFormat...
     D                 PR              *
     D   parser                        *   const
     D   data                     65535A   varying
      *
     D parseLineFormatImage...
     D                 PR              *
     D   formatPtr                     *   const
      *
     D assertConstantLine...
     D                 PR              N   extproc('assertConstantLine')
     D  parserPtr                      *   const
     D  lineformat                     *   const
     D  data                      65535A   varying
      *
     D assertLine...
     D                 PR              N   extproc('assertLine')
     D  parserPtr                      *   const
     D  lineformat                     *   const
     D  data                      65535A   varying
      *
     D isLineValid     PR              N
     D  parser                         *   const
     D  lineFormat                     *   const
     D  data                      65535A   varying
      *
     D getNextLineFormatId...
     D                 PR            10I 0
      *
      /copy 'parser_h.rpgle'
      /copy 'config_provider_h.rpgle'
      /copy 'line_processor_h.rpgle'
      /copy 'validator_h.rpgle'
      /copy 'serializer_h.rpgle'
      /copy 'logger_h.rpgle'
      /copy 'utilities_h.rpgle'
      /copy 'message_h.rpgle'
      /copy 'llist_h.rpgle'
      /copy 'libc_h.rpgle'
      /copy 'rng/input_provider_h.rpgle'


      *---------------------------------------------------------------
      * Dynamic procedure calls
      *---------------------------------------------------------------
     D assertProcPtr   S               *   procptr
     D assert          PR              N   extproc(assertProcPtr)
     D  parserPtr                      *   const
     D  lineformat                     *   const
     D  data                      65535A   varying
      *
      *
     D condProcPtr     S               *   procptr
     D condition_check...
     D                 PR              N   extproc(condProcPtr)
     D  conditionPtr                   *   const
     D  lineFormat                     *   const
     D  data                      10000A   const varying
      *
      *
     D condDisposeProcPtr...
     D                 S               *   procptr
     D condition_dispose...
     D                 PR                  extproc(condDisposeProcPtr)
     D  condPtr                        *


      *---------------------------------------------------------------
      * Variables
      *---------------------------------------------------------------
      /copy 'internals_h.rpgle'


     /**
      * \brief Create parser data structure
      *
      * Creates the parser data structure which is the ground
      * on which every other procedure builds on. It returns a
      * pointer to a data structure which holds all necessary
      * data for parsing.
      *
      * <br><br>
      *
      * The parser is initialized with the following configuration:
      * <ul>
      *   <li>ignore blank lines = true</li>
      *   <li>ignore invalid lines = true</li>
      *   <li>line format matching = value match</li>
      * </ul>
      *
      * \author Mihael Schmidt
      * \date   18.01.2008
      *
      * \return Pointer to parser data structure
      */
     P ffp_create      B                   export
     D                 PI              *
     D parserPtr       S               *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       parserPtr = %alloc(%size(tmpl_parser));
       parser.ignoreBlankLines = *on;
       parser.processInvalidLines = *off;
       parser.lineFormatComplianceLevel = LINE_FORMAT_VALUE_MATCH;
       parser.fieldLengthType = FIELD_LENGTH_TYPE_FIXED;
       parser.lineProcessors = list_create();
       parser.lineformats = list_create();
       parser.logicalLineNumber = 0;
       parser.physicalLineNumber = 0;
       parser.inputProvider = *null;
       parser.configProvider = *null;
       parser.serializer = *null;
       parser.logger = *null;

       return parserPtr;
      /end-free
     P                 E


     /**
      * \brief Set Configuration Provider
      *
      * \author Mihael Schmidt
      * \date   07.04.2010
      *
      * \param Pointer to the parser data structure
      * \param Pointer to the configuration provider
      */
     P ffp_setConfigurationProvider...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
     D   configProvider...
     D                                 *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       parser.configProvider = configProvider;
      /end-free
     P                 E


     /**
      * \brief Parse text file
      *
      * This procedure parses a text file and calls the line processors
      * if the line read is a valid one.
      *
      * \author Mihael Schmidt
      * \date   16.01.2008
      *
      * \param Pointer to parser data structure
      *
      * \throws CPF9898 Input provider not set
      */
     P ffp_parse       B                   export
     D                 PI              N
     D   parserPtr                     *   const
      *
     D len             S             10I 0
     D buffer          S          65535A   varying
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D i               S             10I 0
     D lineFormatPtr   S               *
     D lineProcessor   S               *   based(lineProcessorPtr)
     D valid           S               N   inz(*on)
     D serTransId      S           1000A   varying
     D serBuffer       S                   like(buffer)
     D retVal          S               N   inz(*on)
      /free
       if (parser.inputProvider = *null);
         msg_sendEscapeMessage('Input provider not set.');
       endif;

       // open file
       rng_input_open(parser.inputProvider);

       // document start
       list_abortIteration(parser.lineProcessors);
       lineProcessorPtr = list_getNext(parser.lineProcessors);
       dow (lineProcessorPtr <> *null);
         ffp_processor_process(lineProcessor :
                               parserPtr : DOCUMENT_START : *null : '' : *on);
         lineProcessorPtr = list_getNext(parser.lineProcessors);
       enddo;

       // text einlesen
       dow (rng_input_read(parser.inputProvider : buffer));

         valid = *on;
         parser.logicalLineNumber += 1;
         parser.physicalLineNumber += 1;

         // check if line is empty
         if (parser.ignoreBlankLines and buffer = *blank);
           // blank line => do nothing
         else;

           lineFormatPtr = resolveMatchingLineFormat(parserPtr: buffer);
           if (lineFormatPtr = *null);
             ffp_logger_log(parser.logger :
                            '[' + PROGRAM_ID + '] ' +
                            'There is no matching line format for data row ' +
                            %char(parser.physicalLineNumber) + '.' : buffer);
             valid = *off;

           else;

             // check if the line is valid
             monitor;
               valid = isLineValid(parserPtr : lineFormatPtr : buffer);
               on-error;
                 valid = *off;
                 // TODO log error
             endmon;                    

           endif;

           if (valid or parser.processInvalidLines);
             // send data to every line processor
             lineProcessorPtr = list_getNext(parser.lineProcessors);
             dow (lineProcessorPtr <> *null);
               ffp_processor_process(lineProcessor : parserPtr : LINE :
                                     lineFormatPtr : buffer : valid);
               lineProcessorPtr = list_getNext(parser.lineProcessors);
             enddo;

           elseif (not valid);
             if (parser.serializer <> *null);
               serBuffer = buffer + x'00';
               serTransId = ffp_serializer_serialize(parser.serializer :
                                %addr(buffer : *data));
               ffp_logger_log(parser.logger :'Transaction has been serialized.':
                              serTransId : FFP_LOG_LEVEL_INFO);
             endif;
           endif;

         endif;

         if (not valid);
           retVal = *off;
         endif;

       enddo;

       // document end
       list_abortIteration(parser.lineProcessors);
       lineProcessorPtr = list_getNext(parser.lineProcessors);
       dow (lineProcessorPtr <> *null);
         ffp_processor_process(lineProcessor :
                               parserPtr : DOCUMENT_END : *null : '' : *on);
         lineProcessorPtr = list_getNext(parser.lineProcessors);
       enddo;

       rng_input_close(parser.inputProvider);
       
       return retVal;
      /end-free
     P                 E


     /**
      * \brief Add line processor
      *
      * \author Mihael Schmidt
      * \date   16.01.2008
      *
      * \param Pointer to parser data structure
      * \param Procedure pointer to line processor
      *
      * \info Max. number of line processors is 99.
      */
     P ffp_addLineProcessor...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
     D   processor                     *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D l_processor     S               *
      /free
       l_processor = processor;
       list_add(parser.lineProcessors : %addr(l_processor):%size(l_processor));
      /end-free
     P                 E


     /**
      * \brief Add line format from image
      *
      * This method allows to declare quickly a line format by providing an image.
      *
      * The image symbols indicate the type of the field; fields are separated by a space.
      * <ul>
      *    <li>@ indicates an alphanumeric character</li>
      *    <li># indicates a number </li>
      *    <li>alphanumeric indicates a constant</li>
      * </ul>
      *
      * \author Mihael Schmidt
      * \date 16.01.2008
      *
      * \param Pointer to parser data structure
      * \param Character image string
      *
      * \return pointer to line format or null
      */
     P ffp_addLineFormatFromImage...
     P                 B                   export
     D                 PI              *
     D   parserPtr                     *   const
     D   name                       100A   const varying
     D   chars                         *   const
      *
     D lineFormat      DS                  likeds(ffp_lineFormat) based(ptr)
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       ptr = parseLineFormatImage(chars);
       if (ptr <> *null);
         lineFormat.id = getNextLineFormatId();
         lineFormat.name = name;
         // INFO: the content of the lineFormat will be copied into the list
         //       and not just the pointer to the lineFormat
         list_add(parser.lineFormats : ptr : %size(lineFormat));

         dealloc ptr;
         ptr = list_getLast(parser.lineFormats);
       endif;

       return ptr;
      /end-free
     P                 E


     /**
      * \brief Add a constant line format
      *
      * This method declares a constant character string as a line format.
      * The lines in the document must be exactly the constant line format
      * to be a valid match. Trailing blanks also count.
      *
      * \author Mihael Schmidt
      * \date 18.01.2008
      *
      * \param Pointer to parser data structure
      * \param Constant character string
      *
      * \return pointer to line format or null
      */
     P ffp_addConstantLineFormat...
     P                 B                   export
     D                 PI              *
     D   parserPtr                     *   const
     D   name                       100A   const varying
     D   chars                         *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D lineFormatPtr   S               *
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D length          S             10I 0
      /free
       length = strlen(chars);
       lineFormatPtr = %alloc(%size(ffp_lineFormat));
       clear lineFormat;
       lineFormat.id = getNextLineFormatId();
       lineFormat.assertProcPtr = %paddr('assertConstantLine');
       lineFormat.formatLength = length;
       lineFormat.formatCode = %alloc(length);
       lineFormat.name = name;
       memcpy(lineFormat.formatCode : chars : length);

       list_add(parser.lineFormats : lineFormatPtr : %size(ffp_lineFormat));
       dealloc lineFormatPtr;
       lineFormatPtr = list_getLast(parser.lineFormats);

       return lineFormatPtr;
      /end-free
     P                 E


     /**
      * \brief Dispose of all allocated memory
      *
      * All allocated resources will be freed including line processors,
      * validators, line formats, error handler an serializer.
      *
      * \param Pointer to parser data structure
      */
     P ffp_finalize    B                   export
     D                 PI
     D   parserPtr                     *
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D i               S             10I 0
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D lineProcessor   S               *   based(lineProcessorPtr)
      /free
       // free line formats
       list_abortIteration(parser.lineFormats);

       lineFormatPtr = list_getNext(parser.lineFormats);
       dow (lineFormatPtr <> *null);
         dealloc(n) lineFormat.formatcode;

         // free validators
         for i = 1 to lineFormat.validatorCount;
           ffp_validator_finalize(lineFormat.validators(i));
         endfor;

         lineFormatPtr = list_getNext(parser.lineFormats);
       enddo;

       list_dispose(parser.lineFormats);

       // free logger
       if (parser.logger <> *null);
         ffp_logger_finalize(parser.logger);
       endif;

       // free serializer
       if (parser.serializer <> *null);
         ffp_serializer_finalize(parser.serializer);
       endif;

       // free line processors
       list_abortIteration(parser.lineProcessors);
       lineProcessorPtr = list_getNext(parser.lineProcessors);
       dow (lineProcessorPtr <> *null);
         ffp_processor_finalize(lineProcessor);
         lineProcessorPtr = list_getNext(parser.lineProcessors);
       enddo;

       // free input provider
       if (parser.inputProvider <> *null);
         rng_input_finalize(parser.inputProvider);
       endif;

       // free config provider
       if (parser.configProvider <> *null);
         ffp_config_finalize(parser.configProvider);
       endif;

       // free parser structure
       dealloc(n) parserPtr;
      /end-free
     P                 E


     /**
      * \brief Set parser to ignore blank lines
      *
      * \author Mihael Schmidt
      * \date   23.01.2008
      *
      * \param Pointer to parser data structure
      * \param New value *on/*off
      */
     P ffp_setIgnoreBlankLines...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
     D   value                         N   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       parser.ignoreBlankLines = value;
      /end-free
     P                 E


     /**
      * \brief Set parser to process invalid lines
      *
      * \param Parser
      * \param New value (*on/*off)
      */
     P ffp_setProcessInvalidLines...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
     D   value                         N   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       parser.processInvalidLines = value;
      /end-free
     P                 E


     /**
      * \brief Add field validator
      *
      * Adds the given validator to the passed line format.
      *
      * \param Line format
      * \param Field (1-based)
      * \param Validator
      *
      * \info All registered validators will be finalized on parser
      *       finalization.
      */
     P ffp_addValidator...
     P                 B                   export
     D                 PI
     D   pLineFormat                   *   const
     D   field                       10I 0 const
     D   validator                     *   const
      *
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(pLineFormat)
      /free
       lineFormat.validatorCount += 1;
       lineFormat.validatorField(lineFormat.validatorCount) = field;
       lineFormat.validators(lineFormat.validatorCount) = validator;
      /end-free
     P                 E


     /**
      * \brief Add field to line format
      *
      * Adds a new field defintion to the line format.
      *
      * \author Mihael Schmidt
      * \date   25.01.2008
      *
      * \param Pointer to line format data structure
      * \param Field length
      *
      * \info Max. number of fields per line format is 99.
      */
     P ffp_addLineFormatField...
     P                 B                   export
     D                 PI
     D   lineFormatPtr...
     D                                 *   const
     D   fieldLength                 10I 0 const
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D start           S             10I 0 inz(1)
     D i               S             10I 0
      /free
       // get start pos
       for i = 1 to lineFormat.fieldCount;
         start += lineFormat.fields(i).length;
       endfor;

       lineFormat.fieldCount += 1;
       lineFormat.fields(lineFormat.fieldCount).start = start;
       lineFormat.fields(lineFormat.fieldCount).length = fieldLength;
      /end-free
     P                 E


     /**
      * \brief Set field length type
      *
      * The way to determine the field length will be set here.
      * A valid setting is either the FIELD_LENGTH_TYPE_FIXED,
      * fields have fixed field lengths, or FIELD_LENGTH_TYPE_VAR,
      * fields have variable lengths and are delimited by a character
      * which can declared in xml either at the config element or
      * overriden at the lineformat element.
      *
      * \author Mihael Schmidt
      * \date   16.02.2008
      *
      * \param Pointer to parser data structure
      * \param Type (fixed or variable/delimited)
      */
     P ffp_setFieldLengthType...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
     D   type                        10I 0 const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       parser.fieldLengthType = type;
      /end-free
     P                 E


     /**
      * \brief Load configuration
      *
      * Loads the configuration through the configured configuration provider
      *
      * \param Parser
      *
      * \throws CPF9898 Config provider not set
      */
     P ffp_loadConfiguration...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       if (parser.configProvider = *null);
         msg_sendEscapeMessage('Could not load configuration. Configuration ' +
                               ' provider ist not set.');
       else;
         ffp_config_load(parser.configProvider : parserPtr);
       endif;
      /end-free
     P                 E


     /**
      * \brief Return logical line number
      *
      * Returns the logical line number starting with 1.
      *
      * \param Parser
      *
      * \return Line number or 0 if parsing has not been started
      */
     P ffp_getLogicalLineNumber...
     P                 B                   export
     D                 PI            10I 0
     D   parserPtr                     *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       return parser.logicalLineNumber;
      /end-free
     P                 E


     /**
      * \brief Return physical line number
      *
      * Returns the physical line number starting with 1.
      *
      * \param Parser
      *
      * \return Line number or 0 if parsing has not been started
      */
     P ffp_getPhysicalLineNumber...
     P                 B                   export
     D                 PI            10I 0
     D   parserPtr                     *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       return parser.physicalLineNumber;
      /end-free
     P                 E


     /**
      * \brief Set input provider
      *
      * Sets the input provider for the parser.
      *
      * \param Parser
      * \param Input provider
      */
     P ffp_setInputProvider...
     P                 B                   export
     D                 PI
     D   parserPtr                     *   const
     D   inputProvider...
     D                                 *   const
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
      /free
       parser.inputProvider = inputProvider;
      /end-free
     P                 E


     /**
      * \brief Get line format
      *
      * Returns the line format identified by the passed name.
      *
      * \param Parser
      * \param Line format name
      *
      * \return Line format or *null if there is no line format with the
      *         passed name.
      */
     P ffp_getLineFormat...
     P                 B                   export
     D                 PI              *
     D  parser                         *   const
     D  name                        100A   const varying
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D i               S             10I 0
     D retVal          S               *   inz(*null)
      /free
       list_abortIteration(header.lineFormats);

       lineFormatPtr = list_getNext(header.lineFormats);
       dow (lineFormatPtr <> *null);
         if (lineFormat.name = name);
           list_abortIteration(header.lineFormats);
           return lineFormatPtr;
         endif;

         lineFormatPtr = list_getNext(header.lineFormats);
       enddo;

       return *null;
      /end-free
     P                 E


     /**
      * \brief Set logger
      *
      * \param Parser
      * \param Logger
      */
     P ffp_setLogger...
     P                 B                   export
     D                 PI
     D   parser                        *   const
     D   logger                        *   const
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
      /free
       header.logger = logger;
      /end-free
     P                 E


     /**
      * \brief Get logger
      *
      * \param Parser
      *
      * \return Logger or *null if no logger registered
      */
     P ffp_getLogger...
     P                 B                   export
     D                 PI              *
     D   parser                        *   const
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
      /free
       return header.logger;
      /end-free
     P                 E


     /**
      * \brief Set serializer
      *
      * \param Parser
      * \param Serializer
      */
     P ffp_setSerializer...
     P                 B                   export
     D                 PI
     D   parser                        *   const
     D   serializer                    *   const
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
      /free
       header.serializer = serializer;
      /end-free
     P                 E


     /**
      * \brief Get serializer
      *
      * \param Parser
      *
      * \return Serializer or *null if no serializer registered
      */
     P ffp_getSerializer...
     P                 B                   export
     D                 PI              *
     D   parser                        *   const
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
      /free
       return header.serializer;
      /end-free
     P                 E


     /**
      * \brief Get line formats
      *
      * Returns a list of line formats
      *
      * \param Parser
      *
      * \return List of line formats
      */
     P ffp_getLineFormats...
     P                 B                   export
     D                 PI              *
     D   parser                        *   const
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
      /free
       return header.lineFormats;
      /end-free
     P                 E


     /**
      * \brief Get line processors
      *
      * Returns a list of line processors.
      *
      * \param Parser
      */
     P ffp_getLineProcessors...
     P                 B                   export
     D                 PI              *
     D   parser                        *   const
      *
     D header          DS                  likeds(tmpl_parser) based(parser)
      /free
       return header.lineProcessors;
      /end-free
     P                 E


     /**
      * \brief Execute line validation
      *
      * Checks if the line is valid. The line is valid if all
      * registered validators return valid for the passed data.
      *
      * <br>
      *
      * If no validator is registered the line is also valid.
      *
      * \param Line format
      * \param Data
      *
      * \return *on = line is valid else *off
      */
     P isLineValid     B
     D                 PI              N
     D   parserPtr                     *   const
     D   lineFormatPtr...
     D                                 *   const
     D   data                     65535A   varying
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D validator       S               *
     D valid           S               N
     D rc              S               N
     D i               S             10I 0
     D fieldData       S          65535A   varying
      /free
       for i = 1 to lineFormat.validatorCount;
         fieldData = ffp_util_getLineFormatFieldData(parser.logger :
                         lineFormatPtr : data :
                         lineFormat.validatorField(i));
         if (not ffp_validator_validate(lineFormat.validators(i) : fieldData));
           ffp_logger_log(parser.logger :
                          '[' + PROGRAM_ID + '] ' +
                          'The data row ' +
                          %char(parser.physicalLineNumber) + ' could ' +
                          'not successfully be validated. ' +
                          'Line format: ' + lineFormat.name + '. Field ' +
                          %char(lineFormat.validatorField(i)) + ': ' +
                          fieldData : data);
           return *off;
         endif;
       endfor;

       return *on;
      /end-free
     P                 E


     /**
      * \brief Create line format from character image
      *
      * \author Mihael Schmidt
      * \date   16.01.2008
      *
      * \param Character image string
      */
     P parseLineFormatImage...
     P                 B
     D                 PI              *
     D  formatCode                     *   const
      *
     D lineFormatPtr   S               *
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D length          S             10I 0
      /free
       length = strlen(formatCode);
       lineFormatPtr = %alloc(%size(ffp_lineFormat));
       clear lineFormat;
       lineFormat.assertProcPtr = %paddr('assertLine');
       lineFormat.formatLength = length;
       lineFormat.formatCode = %alloc(length);
       memcpy(lineFormat.formatCode : formatCode : length);
       return lineFormatPtr;
      /end-free
     P                 E


     /**
      * \brief Resolve the matching line format
      *
      * Resolves the matching line format for the resolved data. The first line
      * format that matches the line will be returned.
      *
      * \author Mihael Schmidt
      * \date   16.01.2008
      *
      * \param Parser
      * \param Pointer to null terminated data
      *
      * \return  invalid = *null <br>
      *          valid = pointer to line format for which the data is valid
      */
     P resolveMatchingLineFormat...
     P                 B
     D                 PI              *
     D   parserPtr                     *   const
     D   data                     65535A   varying
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D lineFormatPtr   S               *
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
      /free
       // make certain that we begin the iteration from the beginning
       list_abortIteration(parser.lineFormats);

       lineFormatPtr = list_getNext(parser.lineFormats);
       dow (lineFormatPtr <> *null);

         assertProcPtr = lineFormat.assertProcPtr;
         if (assert(parserPtr : lineFormatPtr : data));
           return lineFormatPtr;
         endif;

         lineFormatPtr = list_getNext(parser.lineFormats);
       enddo;

       list_abortIteration(parser.lineFormats);

       return *null;
      /end-free
     P                 E


     /**
      * \brief Assert constant line
      *
      * Checks if the data is exactly the stored format
      *
      * \author Mihael Schmidt
      * \date   16.01.2008
      *
      * \param Pointer to null terminated data
      *
      * \return *on = data matches format
      *         *off = data does not match format
      */
     P assertConstantLine...
     P                 B
     D                 PI              N
     D  parserPtr                      *   const
     D  lineFormatPtr                  *   const
     D  data                      65535A   varying
      *
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D valid           S               N   inz(*on)
     D length          S             10I 0
      /free
       length = %len(%trimr(data));

       if (length <> lineFormat.formatLength);
         valid = *off;
       elseif (memcmp(lineFormat.formatCode : %addr(data : *DATA) : length)<>0);
         valid = *off;
       endif;

       return valid;
      /end-free
     P                 E


     /**
      * \brief Assert line
      *
      * Checks if the data complies to the given line format.
      * The line format can contain placeholders for numeric and
      * alphanumeric values or constants.
      *
      * \author Mihael Schmidt
      * \date   16.01.2008
      *
      * \param Pointer to null terminated data
      *
      * \return *on = data matches format
      *         *off = data does not match format
      */
     P assertLine...
     P                 B
     D                 PI              N
     D  parserPtr                      *   const
     D  lineFormatPtr                  *   const
     D  data                      65535A   varying
      *
     D parser          DS                  likeds(tmpl_parser) based(parserPtr)
     D lineFormat      DS                  likeds(ffp_lineFormat)
     D                                     based(lineFormatPtr)
     D valid           S               N   inz(*on)
     D length          S             10I 0
     D fPtr            S               *
     D fChar           S              1A   based(fPtr)
     D cPtr            S               *
     D c_ds            DS                  based(cPtr)
     D  char                          1A
     D  cInt                          3I 0
     D i               S             10I 0
     D groupLength     S             10I 0
      /free
       length = %len(%trimr(data));

       if (parser.lineFormatComplianceLevel = LINE_FORMAT_EXACT_MATCH and
           length <> lineFormat.formatLength);
         valid = *off;
       elseif (length > lineFormat.formatLength);
         valid = *off;
       else;

         for i = 0 to lineFormat.formatLength - 1;
           if (not valid);
             leave;
           endif;

           // check if the data length is already reached
           if (i+1 > length);

             fPtr = lineFormat.formatCode + i;

             // only variable field characters (@ or #) may follow
             // after reaching the line limit (and only if not EXACT_MATCH, handled above)
             if (fChar = NUMERIC_CHAR_IMAGE or
                 fChar = ALPHA_CHAR_IMAGE);
               // everything is ok;
             else;
               valid = *off;
             endif;

           else;

             cPtr = %addr(data : *DATA) + i;
             fPtr = lineFormat.formatCode + i;

             select;
               when fChar = NUMERIC_CHAR_IMAGE;
                 if (%check(NUMERIC : char) > 0);
                   valid = *off;
                 endif;

               when fChar = ALPHA_CHAR_IMAGE;


               other; // all other cases must be a constants and so it must be equal
                 if (fChar <> char);
                   valid = *off;
                 endif;

             endsl;

           endif;
         endfor;

       endif;

       return valid;
      /end-free
     P                 E


     /**
      * \brief Get next free line format id
      * 
      * Returns the next line format id.
      * 
      * \return line format id
      */ 
     P getNextLineFormatId...
     P                 B
     D                 PI            10I 0
      *
     D lineFormatId    S             10I 0 static
      /free
       lineFormatId += 1;
       return lineFormatId;
      /end-free
     P                 E
     