     /**
      * \brief Flat File Parser : Range Validator
      *
      * This validator checks if the value is an range of configured values.
      * The range is either expressed by two decimal numbers or two integers. 
      * 
      * \author Mihael Schmidt
      * \date   20.04.2010
      *
      * \implements FFPVAL - Flat File Parser Field Validator
      *
      */

      *-------------------------------------------------------------------------
      *
      * (C) Copyleft 2011 Mihael Schmidt
      *
      * This file is part of Flat File Parser (FFP).
      *
      * FFP is free software: you can redistribute it and/or modify
      * it under the terms of the GNU General Public License as published
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * FFP is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU General Public License for more details.
      *
      * You should have received a copy of the GNU General Public License
      * along with FFP.  If not, see <http://www.gnu.org/licenses/>.
      *
      *-------------------------------------------------------------------------


     H nomain


      *---------------------------------------------------------------
      * Prototypes
      *---------------------------------------------------------------
      /copy 'range_validator_h.rpgle'
      /copy 'logger_h.rpgle'
      /copy 'message_h.rpgle'
      /copy 'libc_h.rpgle'


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
     D FIELD_VALIDATOR_ID...
     D                 C                   'ffp_validator_range'


      *---------------------------------------------------------------
      * Templates
      *---------------------------------------------------------------
     D tmpl_range      DS                  qualified template
     D   von                         30P10
     D   bis                         30P10
      *
      /copy 'validator_t.rpgle'


      *---------------------------------------------------------------
      * Procedures
      *---------------------------------------------------------------

     /**
      * \brief Create range validator
      *
      * Creates a new instance of a range validator.
      *
      * <br><br>
      *
      * The user data must be provide a range of numbers, decimal
      * or integer, which are separated by a space. If it should be
      * a specific value then one value is enough to provide.
      *
      * <br><br>
      *
      * Range example: 100 200<br>
      * Exact value example: 100
      *
      * \param User data (range)
      *
      * \info The lower border must be passed as the first value followed
      *       by the upper border.
      */
     P ffp_validator_range_create...
     P                 B                   export
     D                 PI              *
     D   userdata                      *   const options(*nopass)
      *
     D header          DS                  likeds(ffp_validator)
     D                                     based(validator)
     D range           DS                  likeds(tmpl_range) based(ptr)
     D tmpdata         S           1000A   varying
     D index           S             10I 0
      /free
       validator = %alloc(%size(ffp_validator));

       clear header;

       header.id = FIELD_VALIDATOR_ID;
       header.proc_validate = %paddr('ffp_validator_range_validate');
       header.proc_finalize = %paddr('ffp_validator_range_finalize');
       header.proc_setLogger = %paddr('ffp_validator_range_setLogger');

       if (%parms() = 0 or userdata = *null);
         msg_sendEscapeMessage('A range must be passed to the range ' + 
                               'validator.');
       elseif (strlen(userdata) > 0 and %len(%trim(%str(userdata))) > 0);
         ptr = %alloc(%size(tmpl_range));
         tmpdata = %trim(%str(userdata));

         // wenn nur ein wert bergeben wird, zhlt er als von und bis ( = exakter wert)
         index = %scan(' ' : tmpdata);
         if (index = 0);
           range.von = %dec(tmpdata : 30 : 10);
           range.bis = range.von;
         else;
           range.von = %dec(%subst(tmpdata : 1 : index-1) : 30 : 10);
           range.bis = %dec(%subst(tmpdata : index+1) : 30 : 10);
         endif;

         header.userdata = ptr;
       else;
         msg_sendEscapeMessage('A range must be passed to the range ' + 
                               'validator.');
       endif;

       return validator;
      /end-free
     P                 E


     /**
      * \brief Free resources
      *
      * All previously allocated resource will be freed.
      *
      * \param Field validator
      */
     P ffp_validator_range_finalize...
     P                 B                   export
     D                 PI
     D   validator                     *
      *
     D header          DS                  likeds(ffp_validator)
     D                                     based(validator)
      /free
       if (header.userdata <> *null);
         dealloc header.userdata;
       endif;

       dealloc(n) validator;
      /end-free
     P                 E


     /**
      * \brief Check field value
      *
      * Checks if the field value is in the configured range.
      *
      * \param Field validator
      * \param Data
      */
     P ffp_validator_range_validate...
     P                 B                   export
     D                 PI              N
     D  validator                      *   const
     D  data                      65535A   const varying
      *
     D header          DS                  likeds(ffp_validator)
     D                                     based(validator)
     D length          S             10I 0
     D range           DS                  likeds(tmpl_range)
     D                                     based(header.userdata)
     D retVal          S               N
     D l_data          S             30P10
      /free
       length = %len(%trimr(data));

       monitor;
         l_data = %dec(data : 30 : 10);

         if (range.von <= l_data and
             range.bis >= l_data);
           retVal = *on;
         else;
           ffp_logger_log(header.logger :
               '[' + %trimr(header.id) + '] ' +
               data + ' is out of range ' +
               %char(range.von) + ' - ' + %char(range.bis) + '.');
           retVal = *off;
         endif;

         on-error;
           // no decimal data
           ffp_logger_log(header.logger :
               '[' + %trimr(header.id) + '] ' +
               data + ' is no numeric data.');
           retVal = *off;
       endmon;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Set logger
      *
      * \param Validator
      * \param Logger
      */
     P ffp_validator_range_setLogger...
     P                 B                   export
     D                 PI
     D   validator                     *   const
     D   logger                        *   const
      *
     D header          DS                  likeds(ffp_validator)
     D                                     based(validator)
      /free
       header.logger = logger;
      /end-free
     P                 E

